==> capabilities-0.3.5-py3-none-any.whl: capabilities/core.py <==
import os
import dataclasses
import typing
import dacite
from dataclasses import dataclass, field, is_dataclass
from typing import Dict, Any, List, Type, Union, Literal
from typing import Optional

try:
    from typing import TypeAlias, TypeVar
except:
    from typing_extensions import TypeAlias, TypeVar
import requests
import aiohttp
import asyncio
import time

import aiohttp
import dacite
import requests
from dataclasses import dataclass, field, is_dataclass, asdict
from pydantic import BaseModel
from pydantic.fields import ModelField
from pydantic.main import ModelMetaclass
import logging

logger = logging.getLogger("capabilities")
from typing import Any, Callable, Dict, List, Literal, Optional, Union
from capabilities.config import CONFIG

_CAPABILITIES = {}


@dataclass
class CapabilityBase:
    ...


from termcolor import colored


@dataclass
class Capability(CapabilityBase):
    uri: str
    _capability: Optional[CapabilityBase] = None

    def __call__(self, *args, **kwargs):
        print(
            colored("[  ✓  ]", "green"),
            f"Capability({self.uri}) running with {len(args) + len(kwargs)} arguments",
        )
        return self._capability(*args, **kwargs)

    async def run_async(self, *args, **kwargs):
        print(
            colored("[  ✓  ]", "green"),
            f"Capability({self.uri}) running async with {len(args) + len(kwargs)} arguments",
        )
        return await self._capability.run_async(*args, **kwargs)

    def __post_init__(self):
        try:
            self._capability = _CAPABILITIES[self.uri]
        except KeyError as e:
            print(
                colored("[  ✗  ]", "red"),
                f"KeyError={e}\nCapability lookup failed for uri={self.uri}.\nValid URIs are:",
            )
            for k in _CAPABILITIES.keys():
                print(f"  {k}")


from functools import wraps
import inspect


def register(uri: str) -> Callable:
    """
    A decorator that registers an instance of the decorated class
    under the given URI in _CAPABILITIES.

    Args:
        uri (str): The URI to register the instance of the decorated class.

    Returns:
        Callable: The decorator function
        that receives the decorated class and returns it unmodified.
    """

    def wrapper(cls):
        async def async_run(sync_func, *args, **kwargs):
            loop = asyncio.get_event_loop()
            return await loop.run_in_executor(None, sync_func, *args, **kwargs)

        _CAPABILITIES[uri] = cls() if inspect.isclass(cls) else cls
        call_method = getattr(cls, "__call__")
        if not hasattr(cls, "run_async"):

            @wraps(call_method)
            async def run_async(*args, **kwargs):
                return await async_run(call_method, *args, **kwargs)

            cls.run_async = run_async
        return cls

    return wrapper


@register("blazon/document_qa")
@dataclass
class DocumentQA(CapabilityBase):
    """
    DocumentQA capability that sends the given query to DocumentQA service for answering based on the provided document.

    Attributes:
        None

    Methods:
        __call__(self, document: str, query: str) -> dict:
            Sends the given query to DocumentQA service for answering based on the provided document.
            Args:
                document: A string representing the input document.
                query: A string representing the query for DocumentQA.
            Returns:
                A dictionary containing the answer returned by the DocumentQA service.
            Raises:
                Exception: When the retries hit maximum (8) times and nothing was returned.

        async run_async(self, document: str, query: str, session=None) -> coroutine:
            Sends the given query to DocumentQA service for answering based on the provided document asynchronously.
            Args:
                document: A string representing the input document.
                query: A string representing the query for DocumentQA.
                session: An instance of `aiohttp.ClientSession`.
            Returns:
                A coroutine that resolves to a dictionary containing the answer returned
                by the DocumentQA service.
            Raises:
                Exception: When the retries hit maximum (8) times and nothing was returned.
    """

    def __call__(self, document: str, query: str):
        print(f"[DocumentQA] running query against document with {len(document)} characters")
        return CONFIG.post_sync(
            "/blazon/documentqa", payload={"document": document, "query": query}
        )

    async def run_async(self, document: str, query: str):
        print(f"[DocumentQA] running query against document with {len(document)} characters")
        return CONFIG.post_async(
            "/blazon/documentqa",
            payload={"document": document, "query": query},
        )


@register("blazon/summarize")
@dataclass
class Summarize(CapabilityBase):
    """
    Class for summarizing text using an API call to https://api.blazon.ai/blazon/summarize.

    Args:
        CapabilityBase (class): Base class for all capabilities.

    Methods:
        __call__(self, document: str) -> Dict[str, Any]:
            Method for summarizing `document`. Makes a POST request to the API and returns the JSON response.
            Retries up to 8 times with exponentially increasing sleep times before giving up.
            Args:
                document (str): The text to be summarized.
            Returns:
                Dict[str, Any]: A dictionary object representing the summary, with keys 'summary' (str) and 'score' (float).

        async run_async(self, document: str, session=None) -> Dict[str, Any]:
            Async method for summarizing `document`. Makes an async POST request to the API and returns the JSON response.
            Retries up to 8 times with exponentially increasing sleep times before giving up.
            Args:
                document (str): The text to be summarized.
                session (aiohttp.ClientSession, optional): An aiohttp client session. If not provided, a new one is created. Defaults to None.
            Returns:
                Dict[str, Any]: A dictionary object representing the summary, with keys 'summary' (str) and 'score' (float).
    """

    def __call__(self, document: str):
        print(f"[Summarize] running query against document with {len(document)} characters")
        return CONFIG.post_sync("/blazon/summarize", {"document": document})

    async def run_async(self, document: str):
        print(f"[Summarize] running query against document with {len(document)} characters")
        return await CONFIG.post_async("/blazon/summarize", {"document": document})


StructuredSchema: TypeAlias = Any


def flatten_model(m: Type, path: list[str] = []) -> StructuredSchema:
    """Converts the given type m to a structured schema.

    Args:
        * m (Type): The type to be converted.
        * path (list[str], optional): The path to the current type,
          used for more helpful diagnostic messages saying where
          the conversion failed. Defaults to [].
    """
    orig = typing.get_origin(m)
    if m == list or orig == list:
        args = typing.get_args(m)
        if len(args) != 1:
            p = ".".join(path)
            raise TypeError(
                f"can't deduce list type arg for {m} at {p}, please provide a type annotation to list.",
                path,
            )
        t = flatten_model(args[0])
        return [t]
    elif isinstance(m, ModelMetaclass):
        fields: dict[str, ModelField] = m.__fields__  # type: ignore
        return {k: flatten_model(f.annotation, path=path + [k]) for k, f in fields.items()}
    elif is_dataclass(m):
        return {f.name: flatten_model(f.type, path=path + [f.name]) for f in dataclasses.fields(m)}
    elif m == str:
        return "string"
    elif m == bool:
        return "bool"
    elif m == float:
        return "float"
    elif m == int:
        return "int"
    else:
        p = ".".join(path)
        raise TypeError(
            f"unsupported datatype={m} at {p}\nPlease make sure that all fields are annotated with a bool, int, float, str or a dataclass, list, or pydantic BaseModel. If you need other types supported, let us know!"
        )


def to_dict(obj: Any) -> Any:
    if isinstance(obj, BaseModel):
        return obj.dict()
    elif is_dataclass(obj):
        return dataclasses.asdict(obj)
    elif isinstance(obj, list):
        return [to_dict(o) for o in obj]
    elif isinstance(obj, dict):
        return {k: to_dict(v) for k, v in obj.items()}
    elif isinstance(obj, (str, bool, float, int)):
        return obj
    else:
        raise TypeError(f"unsupported datatype={obj}")


T = TypeVar("T")


def of_dict(t: Type[T], d: Any) -> T:
    if isinstance(t, ModelMetaclass):
        assert isinstance(d, dict)
        return t.parse_obj(d)
    elif is_dataclass(t):
        assert isinstance(d, dict)
        return dacite.from_dict(t, d)  # type: ignore
    elif t == list or typing.get_origin(t) == list:
        assert isinstance(d, list)
        args = typing.get_args(t)
        if len(args) != 1:
            return d  # type: ignore
        (arg,) = args
        return [of_dict(arg, o) for o in d]  # type: ignore
    elif t == dict:
        (kt, vt) = typing.get_args(t)
        assert isinstance(d, dict)
        assert kt in [str, int]
        return {of_dict(kt, k): of_dict(vt, v) for k, v in d.items()}  # type: ignore
    elif t in [str, int, float, bool]:
        assert isinstance(d, t)
        return d
    else:
        raise TypeError(f"unsupported datatype={t}")


def unflatten_model(output_spec, result):
    return (
        output_spec.parse_obj(result)
        if isinstance(output_spec, ModelMetaclass)
        else dacite.from_dict(output_spec, result)
    )


@register("blazon/structured")
@dataclass
class Structured(CapabilityBase):
    """
    `Structured` class allows making requests to the multi API for structured tasks. The class extends CapabilityBase which provides required functionality to interact with multi API. Structured tasks are tasks with specific input and output specs with a natural language instruction.

    Attributes:
        headers (Dict[Any, Any]): HTTP headers to use in requests. It includes the Content-type and API Key parameters.
        url (str): API endpoint URL

    Methods:
        __call__(self, input_spec: ModelMetaclass, output_spec: ModelMetaclass, instructions: str, input: BaseModel) -> Union[output_spec, BaseModel]: Calls the API by sending a payload within a request object. Returns output_spec object if output_spec is ModelMetaclass or if it is an instance of a BaseModel.

        async run_async(self, input_spec: ModelMetaclass, output_spec: ModelMetaclass, instructions: str, input: BaseModel, session=None) -> Union[output_spec, BaseModel]: Calls the API asynchronously. Returns output_spec object if output_spec is ModelMetaclass or if it is an instance of a BaseModel.
    """

    url: str = f"{CONFIG.api_url}/blazon/structured"

    def __call__(
        self,
        input_spec: ModelMetaclass,
        output_spec: ModelMetaclass,
        instructions: str,
        input: Any,
    ):
        payload = dict(
            input_spec=flatten_model(input_spec),
            output_spec=flatten_model(output_spec),
            instructions=instructions,
            input=to_dict(input),
        )
        r = CONFIG.post_sync("/blazon/structured", payload=payload)
        logger.debug("R: ", r)
        result = r["output"]
        return of_dict(output_spec, result)

    async def run_async(
        self,
        input_spec: ModelMetaclass,
        output_spec: ModelMetaclass,
        instructions: str,
        input: BaseModel,
        session=None,
    ):
        payload = dict(
            input_spec=flatten_model(input_spec),
            output_spec=flatten_model(output_spec),
            input=to_dict(input),
            instructions=instructions,
        )
        r = await CONFIG.post_async("/blazon/structured", payload=payload)
        result = r["output"]
        return of_dict(output_spec, result)

==> capabilities-0.3.9-py3-none-any.whl: capabilities/core.py <==
import os
import dataclasses
import typing
import dacite
from dataclasses import dataclass, field, is_dataclass
from typing import Dict, Any, List, Type, Union, Literal
from typing import Optional

try:
    from typing import TypeAlias, TypeVar
except:
    from typing_extensions import TypeAlias, TypeVar
import requests
import aiohttp
import asyncio
import time

import aiohttp
import dacite
import requests
from dataclasses import dataclass, field, is_dataclass, asdict
from pydantic import BaseModel
from pydantic.fields import ModelField
from pydantic.main import ModelMetaclass
import logging

logger = logging.getLogger("capabilities")
from typing import Any, Callable, Dict, List, Literal, Optional, Union
from capabilities.config import CONFIG

_CAPABILITIES = {}


@dataclass
class CapabilityBase:
    ...


from termcolor import colored


@dataclass
class Capability(CapabilityBase):
    uri: str
    _capability: Optional[CapabilityBase] = None

    def __call__(self, *args, **kwargs):
        print(
            colored("[  ✓  ]", "green"),
            f"Capability({self.uri}) running with {len(args) + len(kwargs)} arguments",
        )
        try:
            return self._capability(*args, **kwargs)
        except Exception as e:
            print(
                colored("[  ✗  ]", "red"),
                f"Encountered exception={e}\nRead the docs at https://docs.blazon.ai",
            )
            raise e

    async def run_async(self, *args, **kwargs):
        print(
            colored("[  ✓  ]", "green"),
            f"Capability({self.uri}) running async with {len(args) + len(kwargs)} arguments",
        )
        try:
            return await self._capability.run_async(*args, **kwargs)
        except Exception as e:
            print(
                colored("[  ✗  ]", "red"),
                f"Encountered exception={e}\nRead the docs at https://docs.blazon.ai",
            )
            raise e

    def __post_init__(self):
        try:
            self._capability = _CAPABILITIES[self.uri]
        except KeyError as e:
            print(
                colored("[  ✗  ]", "red"),
                f"KeyError={e}\nCapability lookup failed for uri={self.uri}.\nValid URIs are:",
            )
            for k in _CAPABILITIES.keys():
                print(f"  {k}")


from functools import wraps
import inspect


def register(uri: str) -> Callable:
    """
    A decorator that registers an instance of the decorated class
    under the given URI in _CAPABILITIES.

    Args:
        uri (str): The URI to register the instance of the decorated class.

    Returns:
        Callable: The decorator function
        that receives the decorated class and returns it unmodified.
    """

    def wrapper(cls):
        async def async_run(sync_func, *args, **kwargs):
            loop = asyncio.get_event_loop()
            return await loop.run_in_executor(None, sync_func, *args, **kwargs)

        _CAPABILITIES[uri] = cls() if inspect.isclass(cls) else cls
        call_method = getattr(cls, "__call__")
        if not hasattr(cls, "run_async"):

            @wraps(call_method)
            async def run_async(*args, **kwargs):
                return await async_run(call_method, *args, **kwargs)

            cls.run_async = run_async
        return cls

    return wrapper


@register("blazon/document_qa")
@dataclass
class DocumentQA(CapabilityBase):
    """
    DocumentQA capability that sends the given query to DocumentQA service for answering based on the provided document.

    Attributes:
        None

    Methods:
        __call__(self, document: str, query: str) -> dict:
            Sends the given query to DocumentQA service for answering based on the provided document.
            Args:
                document: A string representing the input document.
                query: A string representing the query for DocumentQA.
            Returns:
                A dictionary containing the answer returned by the DocumentQA service.
            Raises:
                Exception: When the retries hit maximum (8) times and nothing was returned.

        async run_async(self, document: str, query: str, session=None) -> coroutine:
            Sends the given query to DocumentQA service for answering based on the provided document asynchronously.
            Args:
                document: A string representing the input document.
                query: A string representing the query for DocumentQA.
                session: An instance of `aiohttp.ClientSession`.
            Returns:
                A coroutine that resolves to a dictionary containing the answer returned
                by the DocumentQA service.
            Raises:
                Exception: When the retries hit maximum (8) times and nothing was returned.
    """

    def __call__(self, document: str, query: str):
        print(f"[DocumentQA] running query against document with {len(document)} characters")
        return CONFIG.post_sync(
            "/blazon/documentqa", payload={"document": document, "query": query}
        )

    async def run_async(self, document: str, query: str):
        print(f"[DocumentQA] running query against document with {len(document)} characters")
        return await CONFIG.post_async(
            "/blazon/documentqa",
            payload={"document": document, "query": query},
        )


@register("blazon/summarize")
@dataclass
class Summarize(CapabilityBase):
    """
    Class for summarizing text using an API call to https://api.blazon.ai/blazon/summarize.

    Args:
        CapabilityBase (class): Base class for all capabilities.

    Methods:
        __call__(self, document: str) -> Dict[str, Any]:
            Method for summarizing `document`. Makes a POST request to the API and returns the JSON response.
            Retries up to 8 times with exponentially increasing sleep times before giving up.
            Args:
                document (str): The text to be summarized.
            Returns:
                Dict[str, Any]: A dictionary object representing the summary, with keys 'summary' (str) and 'score' (float).

        async run_async(self, document: str, session=None) -> Dict[str, Any]:
            Async method for summarizing `document`. Makes an async POST request to the API and returns the JSON response.
            Retries up to 8 times with exponentially increasing sleep times before giving up.
            Args:
                document (str): The text to be summarized.
                session (aiohttp.ClientSession, optional): An aiohttp client session. If not provided, a new one is created. Defaults to None.
            Returns:
                Dict[str, Any]: A dictionary object representing the summary, with keys 'summary' (str) and 'score' (float).
    """

    def __call__(self, document: str):
        print(f"[Summarize] running query against document with {len(document)} characters")
        return CONFIG.post_sync("/blazon/summarize", {"document": document})

    async def run_async(self, document: str):
        print(f"[Summarize] running query against document with {len(document)} characters")
        return await CONFIG.post_async("/blazon/summarize", {"document": document})


StructuredSchema: TypeAlias = Any


def flatten_model(m: Type, path: list[str] = []) -> StructuredSchema:
    """Converts the given type m to a structured schema.

    Args:
        * m (Type): The type to be converted.
        * path (list[str], optional): The path to the current type,
          used for more helpful diagnostic messages saying where
          the conversion failed. Defaults to [].
    """
    orig = typing.get_origin(m)
    if m == list or orig == list:
        args = typing.get_args(m)
        if len(args) != 1:
            p = ".".join(path)
            raise TypeError(
                f"can't deduce list type arg for {m} at {p}, please provide a type annotation to list.",
                path,
            )
        t = flatten_model(args[0])
        return [t]
    elif isinstance(m, ModelMetaclass):
        fields: dict[str, ModelField] = m.__fields__  # type: ignore
        return {k: flatten_model(f.annotation, path=path + [k]) for k, f in fields.items()}
    elif is_dataclass(m):
        return {f.name: flatten_model(f.type, path=path + [f.name]) for f in dataclasses.fields(m)}
    elif m == str:
        return "string"
    elif m == bool:
        return "bool"
    elif m == float:
        return "float"
    elif m == int:
        return "int"
    else:
        p = ".".join(path)
        raise TypeError(
            f"unsupported datatype={m} at {p}\nPlease make sure that all fields are annotated with a bool, int, float, str or a dataclass, list, or pydantic BaseModel. If you need other types supported, let us know!"
        )


def to_dict(obj: Any) -> Any:
    if isinstance(obj, BaseModel):
        return obj.dict()
    elif is_dataclass(obj):
        return dataclasses.asdict(obj)
    elif isinstance(obj, list):
        return [to_dict(o) for o in obj]
    elif isinstance(obj, dict):
        return {k: to_dict(v) for k, v in obj.items()}
    elif isinstance(obj, (str, bool, float, int)):
        return obj
    else:
        raise TypeError(f"unsupported datatype={obj}")


T = TypeVar("T")


def of_dict(t: Type[T], d: Any) -> T:
    if isinstance(t, ModelMetaclass):
        assert isinstance(d, dict)
        return t.parse_obj(d)
    elif is_dataclass(t):
        assert isinstance(d, dict)
        return dacite.from_dict(t, d)  # type: ignore
    elif t == list or typing.get_origin(t) == list:
        assert isinstance(d, list)
        args = typing.get_args(t)
        if len(args) != 1:
            return d  # type: ignore
        (arg,) = args
        return [of_dict(arg, o) for o in d]  # type: ignore
    elif t == dict:
        (kt, vt) = typing.get_args(t)
        assert isinstance(d, dict)
        assert kt in [str, int]
        return {of_dict(kt, k): of_dict(vt, v) for k, v in d.items()}  # type: ignore
    elif t in [str, int, float, bool]:
        assert isinstance(d, t)
        return d
    else:
        raise TypeError(f"unsupported datatype={t}")


def unflatten_model(output_spec, result):
    return (
        output_spec.parse_obj(result)
        if isinstance(output_spec, ModelMetaclass)
        else dacite.from_dict(output_spec, result)
    )


@register("blazon/structured")
@dataclass
class Structured(CapabilityBase):
    """
    `Structured` class allows making requests to the multi API for structured tasks. The class extends CapabilityBase which provides required functionality to interact with multi API. Structured tasks are tasks with specific input and output specs with a natural language instruction.

    Attributes:
        headers (Dict[Any, Any]): HTTP headers to use in requests. It includes the Content-type and API Key parameters.
        url (str): API endpoint URL

    Methods:
        __call__(self, input_spec: ModelMetaclass, output_spec: ModelMetaclass, instructions: str, input: BaseModel) -> Union[output_spec, BaseModel]: Calls the API by sending a payload within a request object. Returns output_spec object if output_spec is ModelMetaclass or if it is an instance of a BaseModel.

        async run_async(self, input_spec: ModelMetaclass, output_spec: ModelMetaclass, instructions: str, input: BaseModel, session=None) -> Union[output_spec, BaseModel]: Calls the API asynchronously. Returns output_spec object if output_spec is ModelMetaclass or if it is an instance of a BaseModel.
    """

    url: str = f"{CONFIG.api_url}/blazon/structured"

    def __call__(
        self,
        input_spec: ModelMetaclass,
        output_spec: ModelMetaclass,
        instructions: str,
        input: Any,
    ):
        payload = dict(
            input_spec=flatten_model(input_spec),
            output_spec=flatten_model(output_spec),
            instructions=instructions,
            input=to_dict(input),
        )
        r = CONFIG.post_sync("/blazon/structured", payload=payload)
        logger.debug("R: ", r)
        result = r["output"]
        return of_dict(output_spec, result)

    async def run_async(
        self,
        input_spec: ModelMetaclass,
        output_spec: ModelMetaclass,
        instructions: str,
        input: BaseModel,
        session=None,
    ):
        payload = dict(
            input_spec=flatten_model(input_spec),
            output_spec=flatten_model(output_spec),
            input=to_dict(input),
            instructions=instructions,
        )
        r = await CONFIG.post_async("/blazon/structured", payload=payload)
        result = r["output"]
        return of_dict(output_spec, result)
